/* report.hpp --- MicroBenchmark C++ library. */
/*
 * Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
 *
 * This file is part of MicroBenchmark.
 *
 * MicroBenchmark is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * MicroBenchmark is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MicroBenchmark.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#ifndef MICRO_BENCHMARK_CXX_REPORT_HPP
#define MICRO_BENCHMARK_CXX_REPORT_HPP

#include "mbenchmark/c++/c-handler.hpp"
#include "mbenchmark/c++/io.hpp"
#include "mbenchmark/c++/iterator.hpp"

#include <chrono>
#include <cstddef>
#include <cstdio>
#include <vector>

namespace micro_benchmark
{
  MICRO_BENCHMARK_BEGIN_VERSION_NAMESPACE
  namespace detail
  {
    report make_report (void const*, c_suite_ptr const&);
    test_report make_test_report (void const*);
    exec_report make_exec_report (void const*);

    /*  See mbenchmark/stats/time-sample.h  */
    struct time_sample
    {
      bool discarded;
      std::size_t iterations;
      std::chrono::nanoseconds elapsed;
    };

    template <typename T, typename U, typename = void>
    struct unit_base : private U
    {
      using type = T;
      using unit_conversor = U;
      using unit_type = typename U::unit_type;
      using result_type = typename U::result_type;

      unit_base (unit_type u, type v = type{})
        : U (), u (u), v (v)
      { }

      type value () const { return v; }
      result_type to_unit () const { return this->operator() (u, v); }

    private:
      unit_type u;
      type v;
    };

    template <typename T, typename U>
    struct unit_base<T, U, typename U::unit_type>
    {
      using type = T;
      using unit_conversor = U;
      using unit_type = void;
      using result_type = typename U::result_type;

      unit_base (type v = type{}) : v (v) { }
      type value () const { return v; }
      result_type to_unit () const { return v; }

    private:
      type v;
    };

    template <typename T, typename U, int p>
    struct unit;

    template <typename T, typename U>
    struct unit<T, U, 1> : public unit_base<T, U>
    {
      using squared = unit<T, U, 2>;
      using unit_base<T, U>::unit_base;
    };

    template <typename T, typename U>
    struct unit<T, U, 2> : public unit_base<T, U>
    {
      using unit_base<T, U>::unit_base;
    };

    /*  See mbenchmark/stats/time-sample.h  */
    template <typename Unit>
    struct stat_value
    {
      using unit_type = Unit;
      Unit mean;
      Unit std_deviation;
      typename Unit::squared variance;
    };

    struct time_conversor
    {
      using unit_type = int;
      using result_type = std::chrono::nanoseconds;
      result_type operator() (int type, double value) const;
    };

    struct iteration_conversor
    {
      using unit_type = void;
      using result_type = double;
    };

    template <typename Conversor>
    using stat_unit = unit<double, Conversor, 1>;

    using time_stat_value = stat_value<stat_unit<time_conversor>>;
    using iteration_stat_value = stat_value<stat_unit<iteration_conversor>>;
  }

  class exec_report : detail::c_handler
  {
    friend exec_report detail::make_exec_report (void const*);
    explicit exec_report (detail::c_handler&& handler)
      : detail::c_handler (std::move(handler))
    { }

  public:
    char const* name () const;
    std::vector<std::size_t> sizes () const;
    std::size_t iterations () const;
    std::size_t total_iterations () const;
    std::chrono::nanoseconds total_time () const;
    std::vector<detail::time_sample> time_samples () const;
    std::size_t total_samples () const;
    std::size_t used_samples () const;
    detail::time_stat_value iteration_time () const;
    detail::time_stat_value sample_time () const;
    detail::iteration_stat_value sample_iterations () const;

    /* TODO: Meters.  */
  };

  class test_report : detail::c_handler
  {
    using backend = std::vector<exec_report>;
    backend cache;

    friend test_report detail::make_test_report (void const*);
    explicit test_report (detail::c_handler&& handler);

  public:
    using reverse_iterator = backend::const_reverse_iterator;
    using const_iterator = backend::const_iterator;
    using iterator = const_iterator;
    using value_type = iterator::value_type;
    using reference = iterator::reference;
    using pointer = iterator::pointer;

    char const* name () const;

    iterator
    begin () const
    {
      return cache.begin ();
    }

    reverse_iterator
    rbegin () const
    {
      return cache.rbegin ();
    }

    iterator
    end () const
    {
      return cache.end ();
    }

    reverse_iterator
    rend () const
    {
      return cache.rend ();
    }
  };

  class report : detail::c_handler
  {
    using backend = std::vector<test_report>;
    using test_iterator = test_report::iterator;
    using r_test_iterator = test_report::reverse_iterator;

    detail::c_suite_ptr suite;
    backend cache;

    friend report detail::make_report (void const*, detail::c_suite_ptr const&);
    explicit report (detail::c_handler&& handler, detail::c_suite_ptr const&);

  public:
    using reverse_iterator = backend::const_reverse_iterator;
    using const_iterator = backend::const_iterator;
    using iterator = const_iterator;
    using value_type = iterator::value_type;
    using reference = iterator::reference;
    using pointer = iterator::pointer;
    using exec_iterator = iterator_helper::composed<iterator, test_iterator>;
    using exec_range = iterator_helper::range<exec_iterator>;

    using r_exec_iterator = iterator_helper::composed<reverse_iterator,
                                                      r_test_iterator,
                                                      true>;
    using r_exec_range = iterator_helper::range<r_exec_iterator>;

    report () = default;
    report (report&&) = default;
    report& operator= (report&&) = default;
    char const* name () const;
    void print () const;
    void print (io::output_values const& values) const;
    void print (std::FILE* out, io::output_type type) const;
    void print (std::FILE* out, io::output_values const& values) const;
    void print (std::FILE* out, io::output_type type,
                io::output_values const& value) const;

    backend const& tests () const { return cache; }
    bool empty () const { return cache.empty (); }
    std::size_t size () const { return cache.size (); }
    iterator begin () const { return cache.begin (); }
    reverse_iterator rbegin () const { return cache.rbegin (); }
    iterator end () const { return cache.end (); }
    reverse_iterator rend () const { return cache.rend (); }

    exec_range
    executions () const
    {
      using range = iterator_helper::range<iterator>;
      range begin{cache.begin (), cache.end ()};
      range end{cache.end (), cache.end ()};
      return {exec_iterator{begin}, exec_iterator{end}};
    }

    r_exec_range
    reverse_executions () const
    {
      using range = iterator_helper::range<reverse_iterator>;
      range begin{cache.rbegin (), cache.rend()};
      range end{cache.rend (), cache.rend ()};
      return {r_exec_iterator{begin}, r_exec_iterator{end}};
    }
  };
  MICRO_BENCHMARK_END_VERSION_NAMESPACE
}
#endif
