/* test.h --- MicroBenchmark library. */
/*
 * Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
 *
 * This file is part of MicroBenchmark.
 *
 * MicroBenchmark is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * MicroBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MicroBenchmark.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#ifndef MICRO_BENCHMARK_TEST_H
#define MICRO_BENCHMARK_TEST_H

#include "mbenchmark/common.h"
#include "mbenchmark/test/definition.h"
#include "mbenchmark/clock/chrono-provider.h"
#include "mbenchmark/stats/custom.h"

/* *INDENT-OFF* */
MICRO_BENCHMARK_BEGIN_DECLS
/* *INDENT-ON* */

/**
 * Internal definition of the registered test case.
 */
struct micro_benchmark_test_case_definition;

typedef struct micro_benchmark_test_case_definition
  micro_benchmark_test_case_definition;

typedef micro_benchmark_test_case_definition *micro_benchmark_test_case;

typedef struct micro_benchmark_custom_meter
{
  /** Meter used as template.  */
  micro_benchmark_meter meter;
  /** Extra data provided to the meter.  */
  const void *extra;
  /** Data collection associated to the custom meter.  */
  micro_benchmark_custom_sample_collector collector;
} micro_benchmark_custom_meter;

/**
 * Name of the test case.
 *
 * \param test The test case.
 * \return The name provided at registration time.
 */
const char *micro_benchmark_test_case_get_name (micro_benchmark_test_case);

/**
 * Definition of the test case.
 *
 * \param test The test case.
 * \return Definition used for the test case.
 */
micro_benchmark_test
micro_benchmark_test_case_get_definition (micro_benchmark_test_case test);

/**
 * Disable or enable the test case.
 *
 * \param tc The test case.
 * \param e false to skip the test.
 */
void micro_benchmark_test_case_set_enabled (micro_benchmark_test_case tc,
                                            bool e);
/**
 * Provide specific data for the test case.
 *
 * \param tc The test case.
 * \param ptr Data pointer.
 * \param cleanup Function called to release the provided data.
 */
void micro_benchmark_test_case_set_data (micro_benchmark_test_case tc,
                                         void *, void (*)(void *));
/**
 * Data provided by micro_benchmark_test_case_set_data.
 *
 * \param tc The test case.
 */
void *micro_benchmark_test_case_get_data (micro_benchmark_test_case tc);

/**
 * Add a new dimension to the test with the provided sizes.
 *
 * \param tc The test case.
 * \param num_sizes Number of size points on the added dimension.
 * \param sizes Array of size points for the added dimension.
 */
void
micro_benchmark_test_case_add_dimension (micro_benchmark_test_case tc,
                                         size_t num_sizes,
                                         const size_t *sizes);

/**
 * Test parameters.
 * \param tc The test case.
 * \return Number of dimensions of the input provided to the test.
 */
size_t micro_benchmark_test_case_dimensions (micro_benchmark_test_case tc);

/**
 * Test parameters.
 * \param tc The test case.
 * \param dimension The dimension to check.
 * \param n_sizes Output pointer for the number of sizes on this dimension.
 * \param dimension Output pointer for the sizes on this dimension.
 */
void
micro_benchmark_test_case_get_dimension (micro_benchmark_test_case tc,
                                         size_t dimension,
                                         size_t *n_sizes,
                                         const size_t **sizes);

/**
 * Skip the provided iterations before start measuring.
 *
 * \param tc The test case.
 * \param iterations_to_skip Number of iterations not taken into account.
 */
void
micro_benchmark_test_case_skip_iterations (micro_benchmark_test_case tc,
                                           size_t iterations_to_skip);

/**
 * Test limits.
 * \param tc The test case.
 * \return Number of iterations not taken into account.
 */
size_t
micro_benchmark_test_case_iterations_to_skip (micro_benchmark_test_case tc);

/**
 * Limit the total number of iterations performed, not including
 * skipped ones.
 *
 * \param tc The test case.
 * \param min_iterations Minimum number of iterations.
 * \param max_iterations Maximum number of iterations.
 */
void
micro_benchmark_test_case_limit_iterations (micro_benchmark_test_case tc,
                                            size_t min_iterations,
                                            size_t max_iterations);

/**
 * Test limits.
 * \param tc The test case.
 * \return Minimum number of iterations performed.
 */
size_t
micro_benchmark_test_case_min_iterations (micro_benchmark_test_case tc);

/**
 * Test limits.
 * \param tc The test case.
 * \return Maximum number of iterations performed.
 */
size_t
micro_benchmark_test_case_max_iterations (micro_benchmark_test_case tc);

/**
 * Limit the total number of iterations performed on each sample.
 *
 * \param tc The test case.
 * \param min_iterations Minimum number of iterations.
 * \param max_iterations Maximum number of iterations.
 */
void
micro_benchmark_test_case_limit_samples (micro_benchmark_test_case tc,
                                         size_t min_iterations,
                                         size_t max_iterations);

/**
 * Test limits.
 * \param tc The test case.
 * \return Minimum number of iterations per sample.
 */
size_t
micro_benchmark_test_case_min_sample_iterations (micro_benchmark_test_case
                                                 tc);

/**
 * Test limits.
 * \param tc The test case.
 * \return Maximum number of iterations per sample.
 */
size_t
micro_benchmark_test_case_max_sample_iterations (micro_benchmark_test_case
                                                 tc);

/**
 * Maximum execution time of the test code per execution.
 *
 * \param tc The test case.
 * \param d Effective deadline per execution.
 */
void
micro_benchmark_test_case_set_max_time (micro_benchmark_test_case tc,
                                        micro_benchmark_clock_time d);

/**
 * Maximum execution time of the test code per execution.
 *
 * \param tc The test case.
 * \return Value for the test case.
 */
micro_benchmark_clock_time
micro_benchmark_test_case_get_max_time (micro_benchmark_test_case tc);

/**
 * Chronometer used.
 *
 * \param tc The test case.
 * \param type Clock type.
 * \param p Implementation provider.
 */
void
micro_benchmark_test_case_set_chrono (micro_benchmark_test_case tc,
                                      micro_benchmark_clock_type type,
                                      micro_benchmark_chronometer_provider p);

/**
 * Chronometer used.
 *
 * \param tc The test case.
 * \param type Clock type.
 * \param m Custom chronometer.
 */
void
micro_benchmark_test_case_set_custom_chrono (micro_benchmark_test_case tc,
                                             micro_benchmark_clock_type type,
                                             micro_benchmark_meter m);

/**
 * Set custom statistical calculator.
 *
 * \param tc The test case.
 * \param calculator Function called to process the time samples.
 */
void
micro_benchmark_test_case_set_calculator (micro_benchmark_test_case tc,
                                          micro_benchmark_custom_time_calculator);

/**
 * Add custom meter.
 *
 * \param tc The test case.
 * \param m Additional meter used to collect samples.
 */
void
micro_benchmark_test_case_add_meter (micro_benchmark_test_case tc,
                                     micro_benchmark_custom_meter m);

MICRO_BENCHMARK_END_DECLS
#endif
