/* functions.h --- MicroBenchmark library. */
/*
 * Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
 *
 * This file is part of MicroBenchmark.
 *
 * MicroBenchmark is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * MicroBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MicroBenchmark.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#ifndef MICRO_BENCHMARK_UTILITY_FUNCTIONS_H
#define MICRO_BENCHMARK_UTILITY_FUNCTIONS_H

#include "mbenchmark/common.h"
#include "mbenchmark/test.h"

/* *INDENT-OFF* */
MICRO_BENCHMARK_BEGIN_DECLS
/* *INDENT-ON* */

/**
 * \brief Ensure the processing of the provided values.
 *
 * The compiler will happily optimize values that won't be visible to
 * the outside world, which is certainly the case with most of the
 * values generated by the code under test.  Therefore, the test may
 * provide those results to this function in order to avoid the code
 * elimination.
 *
 * Supported compilers will not call this function but inline the
 * expected invocation.
 *
 * \param dummy First parameter for the vararg.
 * \return false if the collected data seems enough, true otherwise.
 */
void micro_benchmark_do_not_optimize (const void *dummy);

/**
 * Initialize library subsystems.
 */
void micro_benchmark_init (void);

/**
 * Release library resources.
 */
void micro_benchmark_cleanup (void);

/**
 * \brief Main function.
 *
 * Parses command line arguments and runs the statically registered
 * tests.
 *
 * \param argc Argument count
 * \param argv Argument vector.
 * \return Exit code for \code{main}.
 */
int micro_benchmark_main (int argc, char **argv);

/**
 * \brief Static registration of the test.
 *
 * \param test Test to register.
 * \return Exit code for the operating system.
 */
void micro_benchmark_register_static_test (const char *name,
                                           micro_benchmark_test test);

/** Static constraint registration prototype.  */
typedef void (*micro_benchmark_static_constraint) (micro_benchmark_test_case);

/**
 * \brief Static registration of a test constraint.
 *
 * \param name Registered name of the test to constraint.
 * \param constraints Function called to constraint the test before
 *        registration.
 */
void
micro_benchmark_register_static_constraint (const char *name,
                                            micro_benchmark_static_constraint);

MICRO_BENCHMARK_END_DECLS
#endif
