/* xalloc.c --- Abort when there is no memory. */
/*
 * Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
 *
 * This file is part of MicroBenchmark.
 *
 * MicroBenchmark is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * MicroBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MicroBenchmark.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "internal/utility/xalloc.h"
#include "internal/utility/error.h"
#include "internal/utility/intl.h"

#include <assert.h>
#include <stdlib.h>
#include <stdint.h>

#if defined(__GNUC__) && defined(__SANITIZE_ADDRESS__)
/* Define the actual functions.  */
#undef xmalloc
#undef xmalloc_n
#undef xcalloc
#undef xcalloc_n
#undef xrealloc_n
#undef xfree
#undef xfree_n
#endif

void *
xmalloc (size_t size)
{
  assert (size > 0);
  void *ret = malloc (size);
  MB_HANDLE_ERROR (!ret, "Error allocating memory\n");
  return ret;
}

void *
xmalloc_n (size_t size, size_t n)
{
  assert (size > 0);
  assert (SIZE_MAX / size > n);
  return xmalloc (size * n);
}

void *
xcalloc_n (size_t size, size_t n)
{
  assert (size > 0);
  void *ret = calloc (n, size);
  MB_HANDLE_ERROR (!ret, "Error allocating memory\n");
  return ret;
}

void *
xcalloc (size_t size)
{
  assert (size > 0);
  return xcalloc_n (size, 1);
}

void *
xrealloc_n (void *ptr, size_t size, size_t n)
{
  assert (size > 0);
  assert (n > 0);
  assert (SIZE_MAX / size > n);
  void *ret = realloc (ptr, size * n);
  MB_HANDLE_ERROR (!ret, "Error reallocating memory\n");
  return ret;
}

void
xfree (void *ptr, size_t size)
{
  assert (ptr);
  assert (size > 0);
  (void) size;
  free (ptr);
}

void
xfree_n (void *ptr, size_t size, size_t n)
{
  assert (ptr);
  assert (size > 0);
  assert (n > 0);
  (void) size;
  (void) n;
  free (ptr);
}
