# mbenchmark.m4 --- Build configuration for MicroBenchmark.
#
# Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
#
# This file is part of MicroBenchmark.
#
# MicroBenchmark is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 3 of the License, or (at
# your option) any later version.
#
# MicroBenchmark is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MicroBenchmark.  If not, see
# <http://www.gnu.org/licenses/>.

# serial 1
AC_DEFUN([MICRO_BENCHMARK_CHECK_TRACES],
 [AC_ARG_ENABLE([traces],
   [AS_HELP_STRING([--enable-traces],
     [Emit log calls of trace level])],
   [micro_benchmark_have_traces=$enableval],
   [micro_benchmark_have_traces=no])

  AC_MSG_CHECKING([if trace messages will be generated by the library])
  AS_IF([test "x$micro_benchmark_have_traces" != xyes],
    [AC_DEFINE([MICRO_BENCHMARK_BUILD_WITHOUT_TRACES], [1],
      [Do not emit trace calls when defined])
     AC_MSG_RESULT([no])],
    [AC_MSG_RESULT([$micro_benchmark_have_traces])])

  AC_ARG_WITH([log-level],
   [AS_HELP_STRING([--with-log-level=LEVEL],
     [Select the default log level (trace, debug, info, warn, error)])],
   [micro_benchmark_default_log_level=$enableval],
   [micro_benchmark_default_log_level=default])

  AC_MSG_CHECKING([for the library log level])
  AC_MSG_CHECKING([$micro_benchmark_default_log_level])
  AS_IF([test "x$micro_benchmark_default_log_level" = xdefault],
    [],
    [test "x$micro_benchmark_default_log_level" = xtrace],
    [AC_DEFINE([MICRO_BENCHMARK_DEFAULT_LOG_LEVEL],
      [MICRO_BENCHMARK_TRACE_LEVEL],
      [Default log level])],
    [test "x$micro_benchmark_default_log_level" = xdebug],
    [AC_DEFINE([MICRO_BENCHMARK_DEFAULT_LOG_LEVEL],
      [MICRO_BENCHMARK_DEBUG_LEVEL],
      [Default log level])],
    [test "x$micro_benchmark_default_log_level" = xinfo],
    [AC_DEFINE([MICRO_BENCHMARK_DEFAULT_LOG_LEVEL],
      [MICRO_BENCHMARK_INFO_LEVEL],
      [Default log level])],
    [test "x$micro_benchmark_default_log_level" = xwarn],
    [AC_DEFINE([MICRO_BENCHMARK_DEFAULT_LOG_LEVEL],
      [MICRO_BENCHMARK_WARNING_LEVEL],
      [Default log level])],
    [test "x$micro_benchmark_default_log_level" = xerror],
    [AC_DEFINE([MICRO_BENCHMARK_DEFAULT_LOG_LEVEL],
      [MICRO_BENCHMARK_ERROR_LEVEL],
      [Default log level])],
    [AC_MSG_ERROR([Log level provided is not valid])])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_COVERAGE],
 [AC_ARG_ENABLE([coverage],
   [AS_HELP_STRING([--enable-coverage],
     [Check library code coverage by the test suite])],
   [micro_benchmark_enable_coverage=$enableval],
   [micro_benchmark_enable_coverage=no])
  AC_MSG_CHECKING([if the code must be compiled with coverage support])
  AC_MSG_RESULT([$micro_benchmark_enable_coverage])
  AS_IF([test "x$micro_benchmark_enable_coverage" != xno],
   [AC_PATH_PROGS([GCOV], [gcov], [gcov-not-found])
    AS_IF([test "x$GCOV" = "xgcov-not-found"],
     [AC_PATH_PROGS([LLVM_COV], [llvm-cov], [llvm-cov-not-found])
      AS_IF([test "x$LLVM_COV" = "xgcov-not-found"],
       [AC_MSG_ERROR([neither gcov nor llvm-cov were found])])
      GCOV="$LLVM_COV gcov"
      AC_SUBST([GCOV], [$GCOV])
     ])
    AC_ARG_VAR([GCOV_FLAGS], [Flags used for gcov])
   ])
  AM_CONDITIONAL([ENABLE_COVERAGE],
   [test "x$micro_benchmark_enable_coverage" = xyes])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_CXX],
 [AC_ARG_ENABLE([c++],
   [AS_HELP_STRING([--disable-c++],
     [Do not compile C++ bindings, even when a C++14+ compiler is available])],
   [micro_benchmark_enable_cxx=$enableval],
   [micro_benchmark_enable_cxx=check])

  AC_MSG_CHECKING([if C++ binding is requested])
  AC_MSG_RESULT([$micro_benchmark_enable_cxx])
  AS_IF([test "x$micro_benchmark_enable_cxx" != xno],
   [AX_CXX_COMPILE_STDCXX([14], [], [optional])
    AS_IF([test "x$HAVE_CXX14" = x0],
     [AS_IF([test "x$micro_benchmark_enable_cxx" != xcheck],
       [AC_MSG_ERROR([C++14 compiler not found])],
       [micro_benchmark_enable_cxx=no])])

    AC_LANG_PUSH([C++])
    AS_IF([test "x$micro_benchmark_enable_cxx" != xno],
     [AC_MSG_CHECKING([inline namespace usability])
      AC_LINK_IFELSE([AC_LANG_PROGRAM([[\
         namespace A { inline namespace B {} }
         namespace A { namespace B { extern bool test; }}
         bool A::test = true;]],
         [[auto v = A::test;]])],
     [AC_MSG_RESULT([yes])],
     [AS_IF([test "x$micro_benchmark_enable_cxx" != xcheck],
       [AC_MSG_FAILURE([C++ compiler does not support inline namespace])],
       [AC_MSG_RESULT([not supported])
        micro_benchmark_enable_cxx=no])])
     ])

    AS_IF([test "x$micro_benchmark_enable_cxx" != xno],
     [AC_MSG_CHECKING([C++ asm extended syntax usability])
       AC_LINK_IFELSE([AC_LANG_PROGRAM([[static int global;]],
          [[int& r = global;
            asm volatile ("" : "+x" (r) :: "memory");]])],
        [AC_MSG_RESULT([yes])
         micro_benchmark_enable_cxx=yes],
        [AS_IF([test "x$micro_benchmark_enable_cxx" != xcheck],
          [AC_MSG_FAILURE([C++ compiler does not support extended asm syntax])],
          [AC_MSG_RESULT([not supported])
           micro_benchmark_enable_cxx=no])])
     ])
    AC_LANG_POP([C++])
   ])
  AM_CONDITIONAL([ENABLE_CXX], [test "x$micro_benchmark_enable_cxx" = xyes])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_GUILE],
 [AC_ARG_ENABLE([guile],
   [AS_HELP_STRING([--disable-guile],
     [Do not compile guile bindings, even when they are available])],
   [micro_benchmark_enable_guile=$enableval],
   [micro_benchmark_enable_guile=check])

  AC_MSG_CHECKING([if Guile binding is requested])
  AC_MSG_RESULT([$micro_benchmark_enable_guile])
  AS_IF([test "x$micro_benchmark_enable_guile" != xno],
   [GUILE_PKG([3.0 2.2]) dnl TODO: Patch this check, don't bail out
    AS_IF([test "x$GUILE_EFFECTIVE_VERSION" = x],
     [AS_IF([test "x$micro_benchmark_enable_guile" != xcheck],
      [AC_MSG_ERROR([Guile not found])],
      [micro_benchmark_enable_guile=no])],
     [GUILE_PROGS
      GUILE_FLAGS
      GUILE_MODULE_AVAILABLE([micro_benchmark_guile_has_exceptions],
       [(ice-9 exceptions)])
      micro_benchmark_exceptions_version=new
      AS_IF([test "x$micro_benchmark_guile_has_exceptions" = xno],
       [micro_benchmark_exceptions_version=old])
      AM_CONDITIONAL([HAS_ICE9_EXCEPTIONS],
       [test "x$micro_benchmark_guile_has_exceptions" = xyes])
      AC_SUBST([EXCEPTIONS_VERSION], [$micro_benchmark_exceptions_version])

      GUILE_SITE_DIR
      AC_MSG_CHECKING([for Guile datarootdir])
      GUILE_DATAROOTDIR=`$PKG_CONFIG --print-errors --variable=datarootdir guile-$GUILE_EFFECTIVE_VERSION`
      AC_MSG_RESULT([$GUILE_DATAROOTDIR])

      AC_MSG_CHECKING([for Guile .scm subdirectory])
      GUILE_SCM_BASE=`echo $GUILE_SITE | $SED -e "s,^$GUILE_DATAROOTDIR/,,"`
      AS_IF([test "x$GUILE_SCM_BASE" = x],
       [AS_IF([test "x$micro_benchmark_enable_guile" != xcheck],
         [AC_MSG_FAILURE([Empty site subdirectory])],
         [AC_MSG_RESULT([empty])
          micro_benchmark_enable_guile=no])],
       [AC_MSG_RESULT([$GUILE_SCM_BASE])])
      AC_SUBST([GUILE_SCM_BASE], [$GUILE_SCM_BASE])

      AC_MSG_CHECKING([for Guile libdir])
      GUILE_LIBDIR=`$PKG_CONFIG --print-errors --variable=libdir guile-$GUILE_EFFECTIVE_VERSION`
      AC_MSG_RESULT([$GUILE_LIBDIR])

      AC_MSG_CHECKING([for Guile .go subdirectory])
      GUILE_GO_BASE=`echo $GUILE_SITE_CCACHE | $SED -e "s,^$GUILE_LIBDIR/,,"`
      AS_IF([test "x$GUILE_GO_BASE" = x],
       [AS_IF([test "x$micro_benchmark_enable_guile" != xcheck],
         [AC_MSG_FAILURE([Empty site-ccache subdirectory])],
         [AC_MSG_RESULT([empty])
          micro_benchmark_enable_guile=no])],
       [AC_MSG_RESULT([$GUILE_GO_BASE])])
      AC_SUBST([GUILE_GO_BASE], [$GUILE_GO_BASE])

      AC_MSG_CHECKING([for Guile extension subdirectory])
      GUILE_EXT_BASE=`echo $GUILE_EXTENSION | $SED -e "s,^$GUILE_LIBDIR/,,"`
      AS_IF([test "x$GUILE_GO_BASE" = x],
       [AS_IF([test "x$micro_benchmark_enable_guile" != xcheck],
         [AC_MSG_FAILURE([Empty extension subdirectory])],
         [AC_MSG_RESULT([empty])
          micro_benchmark_enable_guile=no])],
       [AC_MSG_RESULT([$GUILE_EXT_BASE])])
      AC_SUBST([GUILE_EXT_BASE], [$GUILE_EXT_BASE])

      micro_benchmark_enable_guile=yes
      AC_ARG_VAR([GUILE_FLAGS], [Flags used for Guile execution])
      AC_ARG_VAR([GUILD_FLAGS], [Flags used for Guile compilation])
      AS_IF([test "x$GUILD_FLAGS" = x], [GUILD_FLAGS="-O2"])
     ])
   ])
  AM_CONDITIONAL([ENABLE_GUILE], [test "x$micro_benchmark_enable_guile" = xyes])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_LIBM],
 [oldLIBS=$LIBS
  LIBS=
  LIBM_LIBS=
  AC_SEARCH_LIBS([pow], [m], [LIBM_LIBS=$LIBS], [AC_MSG_ERROR([pow not found])])
  AC_SUBST([LIBM_LIBS], [$LIBM_LIBS])
  LIBS=$oldLIBS
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_GETOPT],
 [AC_CHECK_FUNCS([getopt_long], [], [AC_MSG_ERROR([getopt_long needed])])
  AC_CHECK_HEADERS([getopt.h], [], [AC_MSG_ERROR([getopt.h not found])])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_TIME_TYPES],
 [AC_CHECK_HEADERS([time.h], [], [AC_MSG_ERROR([time.h not found])])

  AC_CHECK_TYPE([struct timespec],
    [micro_benchmark_have_timespec=yes
     AC_DEFINE([HAVE_STRUCT_TIMESPEC], [1], [Use struct timespect for time keeping])],
    [micro_benchmark_have_timespec=no],
    [[#include <time.h>]])
  AM_CONDITIONAL([HAVE_TIMESPEC], [test "x$micro_benchmark_have_timespec" = xyes])

  AC_CHECK_TYPE([struct timeval],
    [micro_benchmark_have_timeval=yes
     AC_DEFINE([HAVE_STRUCT_TIMEVAL], [1], [Use struct timesval for time keeping])],
    [micro_benchmark_have_timeval=no],
    [[#include <sys/time.h>]])
  AM_CONDITIONAL([HAVE_TIMEVAL], [test "x$micro_benchmark_have_timeval" = xyes])

  AS_IF([test "x$micro_benchmark_have_timeval" = xno \
         && test "x$micro_benchmark_have_timespec" = xno],
   [AC_MSG_ERROR([No time type available])])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_CLOCK_TYPES],
 [AC_MSG_CHECKING([for CLOCK_TAI])
  AC_COMPILE_IFELSE(
   [AC_LANG_PROGRAM(
     [[#include <time.h>
     ]],
     [[clockid_t c = CLOCK_TAI]])],
   [AC_DEFINE([HAVE_CLOCK_TAI], [1], [Defined if the system has CLOCK_TAI])
    AC_MSG_RESULT([yes])],
   [AC_MSG_RESULT([no])])

  AC_MSG_CHECKING([for CLOCK_PROCESS_CPUTIME_ID])
  AC_COMPILE_IFELSE(
   [AC_LANG_PROGRAM(
     [[#include <time.h>
     ]],
     [[clockid_t c = CLOCK_PROCESS_CPUTIME_ID]])],
   [AC_DEFINE([HAVE_CLOCK_PROCESS_CPUTIME_ID], [1],
     [Defined if the system has CLOCK_PROCESS_CPUTIME_ID])
    AC_MSG_RESULT([yes])],
   [AC_MSG_RESULT([no])])

  AC_MSG_CHECKING([for CLOCK_THREAD_CPUTIME_ID])
  AC_COMPILE_IFELSE(
   [AC_LANG_PROGRAM(
     [[#include <time.h>
     ]],
     [[clockid_t c = CLOCK_THREAD_CPUTIME_ID]])],
   [AC_DEFINE([HAVE_CLOCK_THREAD_CPUTIME_ID], [1],
     [Defined if the system has CLOCK_THREAD_CPUTIME_ID])
    AC_MSG_RESULT([yes])],
   [AC_MSG_RESULT([no])])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_CLOCK_GETTIME],
 [AC_ARG_ENABLE([clock-gettime],
   [AS_HELP_STRING([--enable-clock-gettime],
     [Enable the clock_gettime timing interface])],
   [micro_benchmark_have_clock_gettime=$enableval],
   [micro_benchmark_have_clock_gettime=check])

  AC_MSG_CHECKING([if clock_gettime is requested])
  AC_MSG_RESULT([$micro_benchmark_have_clock_gettime])

  AS_IF([test "x$micro_benchmark_have_clock_gettime" != xno],
   [oldLIBS=$LIBS
    LIBS=
    AC_SEARCH_LIBS([clock_gettime], [rt],
     [micro_benchmark_have_clock_gettime=yes
      AC_DEFINE([HAVE_CLOCK_GETTIME], [1], [Have clock_gettime function])],
     [AS_IF([test "x$micro_benchmark_have_clock_gettime" != xcheck],
       [AC_MSG_ERROR([clock_gettime was enabled but was not found])],
       [micro_benchmark_have_clock_gettime=no])])

    AC_SUBST([CLOCK_GETTIME_LIBS], [$LIBS])
    LIBS=$oldLIBS
   ])

  AM_CONDITIONAL([HAVE_CLOCK_GETTIME],
    [test "x$micro_benchmark_have_clock_gettime" = xyes])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_GETTIMEOFDAY],
 [AC_ARG_ENABLE([gettimeofday],
   [AS_HELP_STRING([--enable-gettimeofday],
     [Enable the gettimeofday timing interface])],
   [micro_benchmark_have_gettimeofday=$enableval],
   [micro_benchmark_have_gettimeofday=check])

  AC_MSG_CHECKING([if gettimeofday is requested])
  AC_MSG_RESULT([$micro_benchmark_have_gettimeofday])

  AC_CHECK_HEADERS([sys/time.h], [],
     [AS_IF([test "x$micro_benchmark_have_gettimeofday" != xcheck],
       [AC_MSG_ERROR([gettimeofday was enabled but was not found])],
       [micro_benchmark_have_gettimeofday=no])])

  AS_IF([test "x$micro_benchmark_have_gettimeofday" != xno],
   [AC_CHECK_FUNC([gettimeofday],
     [micro_benchmark_have_gettimeofday=yes
      AC_DEFINE([HAVE_GETTIMEOFDAY], [1], [Have gettimeofday function])],
     [AS_IF([test "x$micro_benchmark_have_gettimeofday" != xcheck],
       [AC_MSG_ERROR([gettimeofday was enabled but was not found])],
       [micro_benchmark_have_gettimeofday=no])])])

  AM_CONDITIONAL([HAVE_GETTIMEOFDAY],
    [test "x$micro_benchmark_have_gettimeofday" = xyes])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_TIMER_T],
 [AC_ARG_ENABLE([timer-t],
   [AS_HELP_STRING([--enable-timer-t],
     [Enable the create_timer timing interface])],
   [micro_benchmark_have_timer_t=$enableval],
   [micro_benchmark_have_timer_t=check])

  AC_MSG_CHECKING([if create_timer is requested])
  AC_MSG_RESULT([$micro_benchmark_have_timer_t])

  AS_IF([test "x$micro_benchmark_have_timer_t" != xno],
   [AC_CHECK_HEADERS([signal.h], [],
     [AS_IF([test "x$micro_benchmark_have_timer_t" != xcheck],
       [AC_MSG_ERROR([timer_t was enabled but was not found])],
       [micro_benchmark_have_timer_t=no])])

    oldLIBS=$LIBS
    LIBS=
    AC_SEARCH_LIBS([timer_create], [rt],
     [micro_benchmark_have_timer_t=yes
      AC_DEFINE([HAVE_TIMER_T], [1], [Have per-process timer functions])],
     [AS_IF([test "x$micro_benchmark_have_timer_t" != xcheck],
       [AC_MSG_ERROR([create_timer was enabled but was not found])],
       [micro_benchmark_have_timer_t=no])])

    AC_SUBST([TIMER_T_LIBS], [$LIBS])
    LIBS=$oldLIBS])

  AM_CONDITIONAL([HAVE_TIMER_T],
    [test "x$micro_benchmark_have_timer_t" = xyes])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_TIMERFD],
 [AC_ARG_ENABLE([timerfd],
   [AS_HELP_STRING([--enable-timerfd],
     [Enable the timerfd timing interface])],
   [micro_benchmark_have_timerfd=$enableval],
   [micro_benchmark_have_timerfd=check])

  AC_MSG_CHECKING([if create_timerfd is requested])
  AC_MSG_RESULT([$micro_benchmark_have_timerfd])

  AS_IF([test "x$micro_benchmark_have_timerfd" != xno],
   [AC_CHECK_HEADERS([sys/timerfd.h], [],
     [AS_IF([test "x$micro_benchmark_have_timerfd" != xcheck],
       [AC_MSG_ERROR([timerfd header was not found])],
       [micro_benchmark_have_timerfd=no])])

    AC_CHECK_FUNC([timerfd_create],
     [micro_benchmark_have_timerfd=yes
      AC_DEFINE([HAVE_TIMERFD], [1], [Have timerfd functions])],
     [AS_IF([test "x$micro_benchmark_have_timerfd" != xcheck],
       [AC_MSG_ERROR([timerfd_create was enabled was not found])],
       [micro_benchmark_have_timerfd=no])])
   ])

  AM_CONDITIONAL([HAVE_TIMERFD],
    [test "x$micro_benchmark_have_timerfd" = xyes])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_CLOCK_T],
 [AC_CHECK_TYPE([clock_t], [],
   [AC_MSG_ERROR([clock_t type not found])],
   [[#include <time.h>]])

  AC_CHECK_FUNC([clock], [],
   [AC_MSG_ERROR([clock function not found])])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_TIME_T],
 [AC_CHECK_TYPE([time_t], [],
   [AC_MSG_ERROR([time_t type not found])],
   [[#include <time.h>]])

  AC_CHECK_FUNC([time], [],
   [AC_MSG_ERROR([time function not found])])
 ])

AC_DEFUN([MICRO_BENCHMARK_CHECK_CLOCKS],
 [MICRO_BENCHMARK_CHECK_TIME_TYPES
  MICRO_BENCHMARK_CHECK_CLOCK_TYPES
  MICRO_BENCHMARK_CHECK_CLOCK_GETTIME
  MICRO_BENCHMARK_CHECK_GETTIMEOFDAY
  MICRO_BENCHMARK_CHECK_CLOCK_T
  MICRO_BENCHMARK_CHECK_TIME_T
  MICRO_BENCHMARK_CHECK_TIMER_T
  MICRO_BENCHMARK_CHECK_TIMERFD
 ])
