/* case.c --- MicroBenchmark library. */
/*
 * Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
 *
 * This file is part of MicroBenchmark.
 *
 * MicroBenchmark is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * MicroBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MicroBenchmark.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mbenchmark/test.h"

#include "internal/test.h"
#include "internal/report.h"
#include "internal/utility/log.h"
#include "internal/utility/intl.h"
#include "internal/utility/xalloc.h"

#include <assert.h>
#include <stddef.h>
#include <string.h>

#define MICRO_BENCHMARK_SUBMODULE "case"
#define MICRO_BENCHMARK_MODULE "test:" MICRO_BENCHMARK_SUBMODULE

void
micro_benchmark_test_case_init_ (micro_benchmark_test_case test,
                                 const char *name,
                                 micro_benchmark_test definition)
{
  MB_TRACE ("Init %p with definition %p\n", test, definition);
  assert (test);
  assert (name);
  assert (definition);
  test->name_size = strlen (name) + 1;
  test->name = strdup (name);
  micro_benchmark_test_copy_ (&test->definition, definition);
  micro_benchmark_test_constraints_init_ (&test->constraints);
  test->user_data = NULL;
  test->user_data_release = NULL;
  MB_DEBUG ("Created test case \"%s\"\n", test->name);
}

void
micro_benchmark_test_case_release_ (micro_benchmark_test_case test)
{
  MB_TRACE ("Relase %p\n", test);
  assert (test);
  assert (test->name);
  micro_benchmark_test_constraints_cleanup_ (&test->constraints);
  MB_DEBUG ("Releasing test case \"%s\"\n", test->name);
  xfree (test->name, test->name_size);
  if (test->user_data_release)
    {
      MB_DEBUG ("Releasing user data %p\n", test->user_data);
      test->user_data_release (test->user_data);
    }
}

static mbtest_constraints_ *
get_constraints (micro_benchmark_test_case test)
{
  assert (test);
  return &test->constraints;
}

const mbtest_constraints_ *
micro_benchmark_test_case_constraints_ (micro_benchmark_test_case test)
{
  MB_TRACE ("Retrieving constraints from %p\n", test);
  assert (test);
  return get_constraints (test);
}

const char *
micro_benchmark_test_case_get_name (micro_benchmark_test_case test)
{
  MB_TRACE ("Retrieving name from %p\n", test);
  assert (test);
  return test->name;
}

micro_benchmark_test
micro_benchmark_test_case_get_definition (micro_benchmark_test_case test)
{
  MB_TRACE ("Retrieving definition %p\n", test);
  assert (test);
  return &test->definition;
}

void
micro_benchmark_test_case_set_enabled (micro_benchmark_test_case test, bool e)
{
  MB_TRACE ("Setting enabled %p to %d\n", test, e);
  assert (test);
  test->enabled = e;
}

void
micro_benchmark_test_case_set_data (micro_benchmark_test_case test, void *ptr,
                                    void (*cleanup) (void *))
{
  MB_TRACE ("Setting data of %p to %p\n", test, ptr);
  assert (test);
  assert ((!ptr && !cleanup) || (ptr && cleanup));

  if (test->user_data_release)
    {
      MB_DEBUG ("Releasing old data %p\n", test->user_data);
      test->user_data_release (test->user_data);
    }

  test->user_data = ptr;
  test->user_data_release = cleanup;
}

void *
micro_benchmark_test_case_get_data (micro_benchmark_test_case test)
{
  MB_TRACE ("Getting data of %p\n", test);
  assert (test);
  return test->user_data;
}

void
micro_benchmark_test_case_add_dimension (micro_benchmark_test_case tc,
                                         size_t num_sizes,
                                         const size_t *sizes)
{
  assert (tc);
  micro_benchmark_test_constraints_add_dimension_ (get_constraints (tc),
                                                   num_sizes, sizes);
}

size_t
micro_benchmark_test_case_dimensions (micro_benchmark_test_case tc)
{
  assert (tc);
  return micro_benchmark_test_constraints_dimensions_ (get_constraints (tc));
}

void
micro_benchmark_test_case_get_dimension (micro_benchmark_test_case tc,
                                         size_t dimension,
                                         size_t *n_sizes,
                                         const size_t **sizes)
{
  assert (tc);
  mbtest_constraints_ *c = get_constraints (tc);
  const size_t dsize = micro_benchmark_test_constraints_dimensions_ (c);
  assert (dimension < dsize);
  (void) dsize;
  return micro_benchmark_test_constraints_get_dimension_ (c, dimension,
                                                          n_sizes, sizes);
}

void
micro_benchmark_test_case_skip_iterations (micro_benchmark_test_case tc,
                                           size_t iterations_to_skip)
{
  assert (tc);
  micro_benchmark_test_constraints_skip_iterations_ (get_constraints (tc),
                                                     iterations_to_skip);
}

size_t
micro_benchmark_test_case_iterations_to_skip (micro_benchmark_test_case tc)
{
  assert (tc);
  mbtest_constraints_ *c = get_constraints (tc);
  return micro_benchmark_test_constraints_iterations_to_skip_ (c);
}

void
micro_benchmark_test_case_limit_iterations (micro_benchmark_test_case tc,
                                            size_t min_iterations,
                                            size_t max_iterations)
{
  assert (tc);
  micro_benchmark_test_constraints_limit_iterations_ (get_constraints (tc),
                                                      min_iterations,
                                                      max_iterations);
}

size_t
micro_benchmark_test_case_min_iterations (micro_benchmark_test_case tc)
{
  assert (tc);
  mbtest_constraints_ *c = get_constraints (tc);
  return micro_benchmark_test_constraints_min_iterations_ (c);
}

size_t
micro_benchmark_test_case_max_iterations (micro_benchmark_test_case tc)
{
  assert (tc);
  mbtest_constraints_ *c = get_constraints (tc);
  return micro_benchmark_test_constraints_max_iterations_ (c);
}

void
micro_benchmark_test_case_limit_samples (micro_benchmark_test_case tc,
                                         size_t min_iterations,
                                         size_t max_iterations)
{
  assert (tc);
  micro_benchmark_test_constraints_limit_samples_ (get_constraints (tc),
                                                   min_iterations,
                                                   max_iterations);
}

size_t
micro_benchmark_test_case_min_sample_iterations (micro_benchmark_test_case tc)
{
  assert (tc);
  mbtest_constraints_ *c = get_constraints (tc);
  return micro_benchmark_test_constraints_min_sample_iterations_ (c);
}


size_t
micro_benchmark_test_case_max_sample_iterations (micro_benchmark_test_case tc)
{
  assert (tc);
  mbtest_constraints_ *c = get_constraints (tc);
  return micro_benchmark_test_constraints_max_sample_iterations_ (c);
}

void
micro_benchmark_test_case_set_max_time (micro_benchmark_test_case tc,
                                        micro_benchmark_clock_time d)
{
  assert (tc);
  mbtest_constraints_ *c = get_constraints (tc);
  micro_benchmark_test_constraints_set_max_time_ (c, d);
}
