/* constraints.c --- MicroBenchmark library. */
/*
 * Copyright © 2023 Miguel Ángel Arruga Vivas <rosen644835@gmail.com>
 *
 * This file is part of MicroBenchmark.
 *
 * MicroBenchmark is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * MicroBenchmark is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with MicroBenchmark.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "internal/test/constraints.h"
#include "internal/utility/log.h"
#include "internal/utility/intl.h"
#include "internal/utility/xalloc.h"

#include <assert.h>
#include <stddef.h>
#include <string.h>

#define MICRO_BENCHMARK_SUBMODULE "constraints"
#define MICRO_BENCHMARK_MODULE "test:" MICRO_BENCHMARK_SUBMODULE

static const size_t size_dim = sizeof (mbtest_size_dimension_);

static void
copy_dimension (mbtest_size_dimension_ *lvalue,
                size_t num_sizes, const size_t *sizes)
{
  assert (lvalue);
  assert (sizes);
  lvalue->sizes = xmalloc_n (sizeof (size_t), num_sizes);
  lvalue->num_sizes = num_sizes;
  memcpy (lvalue->sizes, sizes, sizeof (size_t) * num_sizes);
}

void
micro_benchmark_test_constraints_init_ (mbtest_constraints_ *ptr)
{
  MB_TRACE ("Init %p\n", ptr);
  assert (ptr);
  memset (ptr, 0, sizeof (mbtest_constraints_));
}

void
micro_benchmark_test_constraints_cleanup_ (mbtest_constraints_ *ptr)
{
  MB_TRACE ("Cleanup %p\n", ptr);
  assert (ptr);
  if (ptr->size_constraint.has_size)
    {
      for (size_t i = 0; i < ptr->size_constraint.num_dimensions; ++i)
        xfree_n (ptr->size_constraint.dimensions[i].sizes, sizeof (size_t),
                 ptr->size_constraint.dimensions[i].num_sizes);

      xfree_n (ptr->size_constraint.dimensions, size_dim,
               ptr->size_constraint.num_dimensions);
    }

  if (ptr->meter_constraint.has_custom_meters)
    xfree_n (ptr->meter_constraint.custom_meters,
             sizeof (micro_benchmark_custom_meter),
             ptr->meter_constraint.num_custom_meters);
}

void
micro_benchmark_test_constraints_add_dimension_ (mbtest_constraints_ *tc,
                                                 size_t num_sizes,
                                                 const size_t *sizes)
{
  assert (tc);
  const size_t dim = tc->size_constraint.num_dimensions++;
  if (tc->size_constraint.has_size)
    {
      tc->size_constraint.dimensions =
        xrealloc_n (tc->size_constraint.dimensions, size_dim,
                    tc->size_constraint.num_dimensions);
    }
  else
    {
      tc->size_constraint.has_size = true;
      tc->size_constraint.dimensions =
        xmalloc_n (size_dim, tc->size_constraint.num_dimensions);
    }
  copy_dimension (&tc->size_constraint.dimensions[dim], num_sizes, sizes);
}

size_t
micro_benchmark_test_constraints_dimensions_ (const mbtest_constraints_ *tc)
{
  assert (tc);
  return tc->size_constraint.num_dimensions;
}

void
micro_benchmark_test_constraints_get_dimension_ (const mbtest_constraints_
                                                 *tc, size_t dimension,
                                                 size_t *n_sizes,
                                                 const size_t **sizes)
{
  assert (tc);
  assert (n_sizes);
  assert (sizes);
  assert (dimension < tc->size_constraint.num_dimensions);
  *n_sizes = tc->size_constraint.dimensions[dimension].num_sizes;
  *sizes = tc->size_constraint.dimensions[dimension].sizes;
}

void
micro_benchmark_test_constraints_skip_iterations_ (mbtest_constraints_ *tc,
                                                   size_t iterations_to_skip)
{
  assert (tc);
  tc->iteration_constraint.iterations_to_skip = iterations_to_skip;
}

size_t
micro_benchmark_test_constraints_iterations_to_skip_ (const
                                                      mbtest_constraints_ *tc)
{
  assert (tc);
  return tc->iteration_constraint.iterations_to_skip;
}

void
micro_benchmark_test_constraints_limit_iterations_ (mbtest_constraints_ *tc,
                                                    size_t min_iterations,
                                                    size_t max_iterations)
{
  assert (tc);
  tc->iteration_constraint.has_iterations = true;
  tc->iteration_constraint.min_iterations = min_iterations;
  tc->iteration_constraint.max_iterations = max_iterations;
}

size_t
micro_benchmark_test_constraints_min_iterations_ (const mbtest_constraints_
                                                  *tc)
{
  assert (tc);
  return tc->iteration_constraint.min_iterations;
}

size_t
micro_benchmark_test_constraints_max_iterations_ (const mbtest_constraints_
                                                  *tc)
{
  assert (tc);
  return tc->iteration_constraint.max_iterations;
}

void
micro_benchmark_test_constraints_limit_samples_ (mbtest_constraints_ *tc,
                                                 size_t min_iterations,
                                                 size_t max_iterations)
{
  assert (tc);
  tc->iteration_constraint.has_sample_iterations = true;
  tc->iteration_constraint.min_sample_iterations = min_iterations;
  tc->iteration_constraint.max_sample_iterations = max_iterations;
}

size_t
micro_benchmark_test_constraints_min_sample_iterations_ (const
                                                         mbtest_constraints_
                                                         *tc)
{
  assert (tc);
  return tc->iteration_constraint.min_sample_iterations;
}


size_t
micro_benchmark_test_constraints_max_sample_iterations_ (const
                                                         mbtest_constraints_
                                                         *tc)
{
  assert (tc);
  return tc->iteration_constraint.max_sample_iterations;
}

void
micro_benchmark_test_constraints_set_max_time_ (mbtest_constraints_ *tc,
                                                micro_benchmark_clock_time d)
{
  assert (tc);
  tc->time_constraint.has_deadline = true;
  tc->time_constraint.deadline = d;
}


micro_benchmark_clock_time
micro_benchmark_test_constraints_get_max_time_ (const mbtest_constraints_ *tc)
{
  assert (tc);
  return tc->time_constraint.deadline;
}

void
micro_benchmark_test_constraints_set_chrono_ (mbtest_constraints_ *tc,
                                              micro_benchmark_clock_type type,
                                              micro_benchmark_chronometer_provider
                                              p)
{
  assert (tc);
  tc->meter_constraint.has_chronometer = true;
  tc->meter_constraint.chrono_type = type;
  tc->meter_constraint.chrono_provider = p;
  tc->meter_constraint.chrono_tmpl = NULL;
}

void
micro_benchmark_test_constraints_set_custom_chrono_ (mbtest_constraints_ *tc,
                                                     micro_benchmark_clock_type
                                                     type,
                                                     micro_benchmark_meter m)
{
  assert (tc);
  tc->meter_constraint.has_chronometer = true;
  tc->meter_constraint.chrono_type = type;
  tc->meter_constraint.chrono_provider =
    MICRO_BENCHMARK_CHRONO_PROVIDER_USER_PROVIDED;
  tc->meter_constraint.chrono_tmpl = m;
}

void
micro_benchmark_test_constraints_set_calculator_ (mbtest_constraints_ *tc,
                                                  micro_benchmark_custom_time_calculator
                                                  c)
{
  assert (tc);
  tc->meter_constraint.has_custom_stats = true;
  tc->meter_constraint.custom_stats = c;
}

void
micro_benchmark_test_constraints_add_meter_ (mbtest_constraints_ *tc,
                                             micro_benchmark_custom_meter m)
{
  assert (tc);
  const size_t pos = tc->meter_constraint.num_custom_meters++;
  tc->meter_constraint.custom_meters =
    xrealloc_n (tc->meter_constraint.custom_meters,
                sizeof (micro_benchmark_custom_meter),
                tc->meter_constraint.num_custom_meters);
  tc->meter_constraint.has_custom_meters = true;
  tc->meter_constraint.custom_meters[pos] = m;
}
